# Author: Toby Dickenson, <htrd90@zepler.org>
#
# Copyright (c) 1999 Toby Dickenson
#
# Permission to use this software in any way is granted without
# fee, provided that the copyright notice above appears in all
# copies. This software is provided "as is" without any warranty.
#

"""NT User folder product

Authenticates off NT security as applied to a registry key

"""

import string, traceback, time
import win32api, win32con, win32security, win32net, win32netcon
from Acquisition import Implicit
from ImageFile import ImageFile
from AccessControl.Role import RoleManager
from AccessControl.User import BasicUser, BasicUserFolder
from Globals import HTMLFile, Persistent, default__class_init__
from App.Dialogs import MessageDialog

key_prefix = "SOFTWARE\\Digital Creations\\zope\\Users"

_users_root = win32api.RegCreateKey(win32con.HKEY_LOCAL_MACHINE,key_prefix)

maximum_number_of_roles = 100

DOMAIN, DOMAIN_CONTROLLER = ".", None

## Experimental support for authenticating from a domain
## (rather than the local computer) Use this at your own
## risk. If you do, please let me know how you get on. I will
## only include a user interface for these parameters once I
## have independant verification that there are no problems.
##
## Your machine must be a member of the domain, and the user
## who is running the service must have the 'Access the
## Network' permission: This excludes the Local System account.
##
## Uncomment the line below, and substitute your domain name
## and the name of a domain controller.
##
## DOMAIN, DOMAIN_CONTROLLER = 'domainname', r'\\controller'

class NTUserFolder(BasicUserFolder):
    """User folder for NT security"""

    meta_type = 'NTUserFolder'
    id        = 'acl_users'
    title     = 'NTUserFolder'

    manage_options=(
    {'label':'Contents',   'action':'manage_main'},
    {'label':'Security',   'action':'manage_access'},
    {'label':'Undo',       'action':'manage_UndoForm'},
    )

    __ac_permissions__=BasicUserFolder.__ac_permissions__+ (
    ('Manage NT User Folder',   ['manage_edit']),   )

    manage_main=HTMLFile('mainUser', globals())

    def __init__(self, key):
        self.key = key

    def manage_edit(self,key,REQUEST):
        "For changing key in management screen"
        if not valid_key(key):  return MessageDialog(
                                           title  ='Bad Key',
                                           message='That key is invalid',
                                           action ='manage_main')
        self.key=key
        return MessageDialog(
                   title  ='Key Changed',
                   message='That key has been changed',
                   action ='manage_main')

    def getUserNames(self):
        "Returns a list of user names"
        names = []
        resume = 0
        while 1:
            users,junk,resume = win32net.NetUserEnum(DOMAIN_CONTROLLER,10,win32netcon.FILTER_NORMAL_ACCOUNT,resume)
            names = names+map(lambda user: user['name'], users)
            if not resume:
                break
        return names

    def getUsers(self):
        "Return a list of user objects"
        # TODO: see where this is used, how essential is it
        # TODO: possibly enumerate all NT users
        return []

    def getUser(self, name):
        "Return the named user object or None if no such user exists"
        try:
            info=win32net.NetUserGetInfo(DOMAIN_CONTROLLER,name,2)
            name=str(info['name'])
            full_name=str(info['full_name'])
            description=str(info['comment'])
        except win32api.error:
            return None
        return NTUser(name,full_name,description,key_prefix+"\\"+self.key)

    def defined_roles(self):
        "A list of the roles that are defined by this key"
        roles=[]
        hkey=win32api.RegOpenKey(_users_root,self.key)
        try:
            for i in range(maximum_number_of_roles):
                try:
                    roles.append(win32api.RegEnumKey(hkey,i))
                except win32api.error:
                    # No more subkeys
                    break
        finally:
            win32api.RegCloseKey(hkey)
        return roles

    def list_sub(self,a,b):
        "Find all the elements of list a not in b"
        return filter(lambda element, b=b[:]: not(element in b) , a[:])

default__class_init__(NTUserFolder)


class NTUser(BasicUser):
    "An object representing an NT user"

    def __init__(self,name,full_name,description,key):
        self._name = name
        self._full_name = full_name
        self._description = description
        self._elapsed_time = 0
        self._key = key
        self._password = ""
        self._authentic = None
        self._roles = None
        self._token = None
        self._hkey = None

    def __del__(self):
        if self._token is not None:
            win32api.CloseHandle(self._token)
            self._token = None
        if self._hkey is not None:
            win32api.RegCloseKey(self._hkey)
            self._hkey = None

    def getUserName(self):
        "Return the username of a user"
        return self._name

    def getFullUserName(self):
        "Return the full name of a user"
        return self._full_name

    def getDescription(self):
        "Return the description of a user"
        return self._description

    def authenticate(self, password, request):
        if self._authentic is None:
            start = time.clock()
            try:
                # User must be able to log in
                token=win32security.LogonUser(self._name,DOMAIN,password,win32security.LOGON32_LOGON_NETWORK,win32security.LOGON32_PROVIDER_DEFAULT)
                win32security.ImpersonateLoggedOnUser(token)
                try:
                    hkey=win32api.RegOpenKey(win32con.HKEY_LOCAL_MACHINE,self._key)
                    # User must be able to read from the controlling key
                    win32api.RegQueryValue(hkey,None)
                finally:
                    win32security.RevertToSelf()
                self._token = token
                self._hkey = hkey
                self._password = password
                self._authentic = 1
            except win32api.error,e:
                #import traceback,sys
                #apply(traceback.print_exception,sys.exc_info())
                self._authentic = 0
            self._elapsed_time = time.clock()-start
        return self._authentic

    def getRoles(self):
        "Return the list of roles assigned to a user."
        if self._authentic is None:
            return []
        if self._roles is None:
            self._roles = []
            win32security.ImpersonateLoggedOnUser(self._token)
            try:
                for i in range(maximum_number_of_roles):
                    try:
                        role_name=win32api.RegEnumKey(self._hkey,i)
                    except win32api.error:
                        # No more subkeys
                        break
                    try:
                        win32api.RegQueryValue(self._hkey,role_name)
                        self._roles.append(role_name)
                    except win32api.error:
                        # User does not have this role
                        pass
            finally:
                win32security.RevertToSelf()
        return self._roles

    def getTimeToAuthenticate(self):
        "Return the time taken to perform authentication."
        return self._elapsed_time




# Probably over cautious
safe_characters=string.uppercase+string.lowercase+string.digits

def valid_key(key):
    # Must be sensble length
    if not 1<len(key)<50:
        return 0
    # And alphanumeric
    for c in str(key):
        if not c in safe_characters:
            return 0
    # And createable or already created
    try: hkey=win32api.RegCreateKey(_users_root,key)
    except: return 0
    win32api.RegCloseKey(hkey)
    return 1
    


manage_addNTUserFolderForm=HTMLFile('addNTUserFolder', globals())

def manage_addNTUserFolder(self, key, REQUEST):
    """Add NT User folder"""

    if not valid_key(key):
        return MessageDialog(
                   title  ='Bad Key',
                   message='That key is invalid',
                   action ='manage_main')

    try:    self._setObject('acl_users', NTUserFolder(key))
    except: return MessageDialog(
                   title  ='Item Exists',
                   message='This object already contains a User Folder',
                   action ='manage_main')

    self.__allow_groups__=self.acl_users

    return self.manage_main(self,REQUEST,update_menu=1)

